
#include "BasicSolver.hpp"
#include <algorithm>
#include "Logging.hpp"
#include "SimplyFire.hpp"

const char*  basicsolver_cpp_version(){return "$Revision: 1.40 $";}


BasicSolver::BasicSolver(TGlobals& globals):
	glob						(globals),
	//next_ext_var				(FIX_VAR),
	var_act_comp				(),
    var_act_hp               ( glob.sets.var_max_act , glob.sets.var_decay, var_act_comp),
    ass_info					(vals,var_act_hp,glob.sets.bump_agil),
    watching					(vals,glob),
    //watches						(c_info),
    queue_begin                 ( 0     ),
    simplify_assign_idx         ( 0     ),
    simpl_next					( 0     ),
    big_simpl_next				( 0     ),
    did_big_simp				( false),
    lbd_total					(0)
{
	reserveForVariables(0);
	decide(TRUE_LIT);
	asrt(vals.litTrue(TRUE_LIT));
}



BasicSolver::~BasicSolver(){}



TrBool BasicSolver::isSAT(){
	asrt(vals.litTrue(TRUE_LIT));
	if (!glob.isValid())          return FALSE_TrBool;

	//next_ext_var			= Variable(ass_info.varsCount());
    //var_act_comp.ext_border= next_ext_var; LOG1("First extended is " << next_ext_var.toDIMACS());


    lbd_total					= 0;
    simpl_next   			= glob.stats.litsCount();
    simplify_assign_idx		= assign_stck.size();
    big_simpl_next			= glob.sets.simpl_freq > 0 ? simpl_next * glob.sets.simpl_freq : max64;

    if(glob.sets.restart_type == glob.sets.RSTRT_LUBY){
    		luby_restarts		.setScale(glob.sets.lbd_rest_buffr);
			lbd_history		.resize(1);}
    else{
    		lbd_history		.resize(glob.sets.lbd_rest_buffr);}
//glucose
    uint32_t max_l  = glob.stats.cnt_lcls / 3;
    if(max_l  <glob.sets.max_learnts) {
    	glob.sets.max_learnts = max_l >> 1;
    	if(glob.sets.max_learnts < 5000) glob.sets.max_learnts = 5000;
    }
    LOG1("Max learnts limit is "<< glob.sets.max_learnts << " incremented by " << glob.sets.max_learnts_inc);


    //TrBool status = glob.sets.do_simpl  ? preProcess() : UNKNOWN_TrBool;
    TrBool status = UNKNOWN_TrBool;

   	while(glob.stats.cpuUsrTime() <= glob.sets.timeout){asrt(glob.isValid());
		uint32_t cfl_mx 			= luby_restarts.nextRestartGetConflictCount();
		status 					= search(cfl_mx);

    	if(glob.sets.print_console){	glob.stats.printProgress();}

    	if(status.isKnown()) 	break;

    	backjumpKeepLevel(0);
   	}

    return status;
}




ClsP BasicSolver::clsCreate(Literal* lits, uint32_t sz, const bool learnt,const uint32_t& lbd){

    asrt(sz > 2);

    ClsP loc	= glob.clsCreate(lits,sz,learnt,lbd);

    watching	.clsAddWtched(lits[0],lits[1],loc);
    watching	.clsAddWtched(lits[1],lits[0],loc);

    cnt_stat(lcls ,1);
    //asrt(glob.c_info.validEndData(loc)); asrt(glob.c_info.validEndData(loc+Cls::allocSizeFor(sz)-1));

    return loc;
}


bool BasicSolver::addClause(vector<Literal>& lits){

    asrt(decisionLevel() == 0);

    if (!glob.isValid() || lits.empty()){								return glob.setInvalid("Added empty clause");}

    // Check if clause is satisfied and remove false/duplicate literals:
    std::sort(lits.begin(),lits.end());
    reserveForVariables(lits.back().toVar().idx());

    Literal                     last    (INVALID_LIT);
    uint32_t                    i       (0);
    vector<Literal>::iterator   to      (lits.begin());
    while(i < lits.size()){
        if(vals.litTrue(lits[i]) || lits[i] == last.flipped()){       		return true;}


        if(vals.litUkwn(lits[i]) && lits[i] != last){
           *to++ = last = lits[i];}

        ++i;
    }

    return uncheckedAddClause(lits.begin(),to);
}




bool BasicSolver::uncheckedAddClause  (vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend){


    uint32_t l_sz(lend-lbeg);

    if(l_sz == 0){										return glob.setInvalid("Added (unchecked) empty clause");}

    if(l_sz == 1){
        asrt(vals.litUkwn(*lbeg));
        decide(*lbeg);		                         	return (bcp() == BCP_SUCCESS);}

    if(l_sz == 2){
    	watching.clsCreateBinary(*lbeg,*(lbeg+1));			return true;}


    clsCreate(lbeg.base(),l_sz,false,l_sz<=BDReason::LBD_MAX?l_sz:BDReason::LBD_MAX);
    return true;
}


TrBool BasicSolver::search(uint32_t max_confl ){

    asrt(glob.isValid());
    TrBool   result	= UNKNOWN_TrBool;
    uint32_t 	conflc	= 0;
    bool restart;
    //vector<Literal> decisions;

    cnt_stat(strts,1)

    while(true ){

        int32_t bcp_res    =     bcp();

        if(bcp_res == BCP_SUCCESS){         //there is no conflict

        	restart = (glob.sets.restart_type == glob.sets.RSTRT_LUBY) ?
        			conflc >= max_confl : //luby
					lbd_history.enoughData()&& ((lbd_history.getAvg()*0.7) > (lbd_total / (glob.stats.cnt_confl+1))); //glucose

        	if(restart) {
        		LOG2("Clear Lbd History");
        		lbd_history.clear();
				if (ass_info.lessAgilThan(2500000)){ break;} //Biere
				inc_stat(rstrts_skip,1);
        	}

        	//if(curr_long_lrnts > glob.sets.max_learnts){learntsGC(); }
        	if(glob.stats.cnt_confl >= (glob.stats.cnt_gc+1)* glob.sets.max_learnts){
        		LOG2("Start Learnts GC");
        		learntsGC();
        	}

             // Simplify the set of problem clauses
            if (decisionLevel() == 0    && (result = simplify(), result.isKnown())) {    return result;}

            // New decision:
            Literal next    = nextDecision();

            //check if model is found
            if (next == INVALID_LIT)    {return TRUE_TrBool;}

            // Increase decision level and enqueue 'next'
            asrt(queue_begin == assign_stck.size()); asrt(vals.litUkwn(next));
            LOG2("Next decision lit " << next.toDIMACS());
            newDecisionLevel();
            decide(next);


        }else{                         //there is a conflict

        	asrt(bcp_res != 1);
			if (!firstUIP(bcp_res))  {                                                             	return FALSE_TrBool; }
			asrt(ass_info.litAssLev(lemma[0]) == decisionLevel());
			asrt(!var_flags.isSet(lemma[0].toVar())); asrt(!isSetLevel(decisionLevel())); asrt(!isSetLevel(0));

			//glucose
			lbd_history	.keep(confl_lbd);
			lbd_total	+= confl_lbd;

			if(confl_lbd >= BDReason::LBD_MAX){
				confl_lbd=BDReason::LBD_MAX;

			}else if(glob.sets.bump_glucose){
				for(uint32_t i=conflVars.size();i--;){
					const BDReason& ri = ass_info.varAssignReason(conflVars[i]); asrt(vals.varKnwn(conflVars[i]));//asrt(ri.isClsReason());

					if(ri.getLBD() < confl_lbd){
						asrt(ri.isClsReason());aux_var(Cls& clr = glob.getCls(ri.getClsReason());)	asrt(clr.isLrnt()); asrt(clr.compLBD() < confl_lbd);
						varBump(conflVars[i]);
					}else{
						asrt(!ri.isClsReason() || !glob.getCls(ri.getClsReason()).isLrnt() || glob.getCls(ri.getClsReason()).compLBD() >= confl_lbd);
					}
				}}


			reduceSelfImpl();

			uint32_t btrk_lev = backtrackLevel(); asrt(btrk_lev < decisionLevel());

			inc_stat(sum_bck_jmp,decisionLevel() - btrk_lev);
			inc_stat(sum_confl_depth,decisionLevel());
			//statistics.storeBackjumpingDistance(bj_dist);


			/*decisions.clear();
			for(uint32_t d=btrk_lev;d!=dec_lev_end.size();++d){
				decisions.push_back(assign_stck[dec_lev_end[d]]);
			}*/

			backjumpKeepLevel(btrk_lev);
			asrt(vals.litUkwn(lemma[0]));

			var_act_hp.decayAll();

			const uint32_t l_sz(lemma.size());


			++conflc;
			if(l_sz > 2){
				asrt(confl_lbd >1);

				glob.registrLBD(confl_lbd);
				ClsP learnt_cl = clsCreate(lemma.begin().base(),l_sz,true,confl_lbd);
				enqueue(lemma[0], learnt_cl,confl_lbd);
				asrt(glob.getCls(learnt_cl).isLrnt()); asrt(glob.getCls(learnt_cl).compLBD() == confl_lbd); asrt(confl_lbd == ass_info.reasonLBD(lemma[0].toVar()));

				//clsBump(learnt_cl);


			}else if(l_sz == 2){
				watching.clsCreateBinary(lemma[0],lemma[1]);

				asrt(vals.litFalse(lemma[1])); 
				enqueue(lemma[0],lemma[1]); asrt(ass_info.reasonLBD(lemma[0].toVar()) == BDReason::LBD_MAX);
			}else{
				decide(lemma[0]); asrt(ass_info.reasonLBD(lemma[0].toVar()) == BDReason::LBD_MAX);
			}
        }
    }

    //maximal number of conflicts reached
    return UNKNOWN_TrBool;

}

void    BasicSolver::backjumpUndoAssigns (uint32_t assign_idx){

	Variable x = INVALID_VAR;
	asrt(assign_stck.size() >= assign_idx);

	queue_begin 	= assign_stck.size();

	while(queue_begin != assign_idx ){
		x  = assign_stck[--queue_begin].toVar();
		vals.varClearValue(x);
		var_act_hp.restoreElementIf(x);}

	assign_stck.resize(queue_begin);
}

// this is only a test
/*
Literal BasicSolver::nextDecisionStrict(){
	if(glob.sets.pol_mode <= 2) return nextDecision();
	bool pol 	= (glob.sets.pol_mode - 3); //3 is force false, 4 is force true
	bool rand	= randy.nextInt(1000) < glob.sets.random_var_freq_permil;

	vector<Literal> bad_pol;
	Literal ret = INVALID_LIT;
	while(true){
		ret	= nextDecision(rand,2);
		if(ret == INVALID_LIT || (ret.isPos()==pol)) break;
		bad_pol.push_back(ret);
	}

	uint32_t strt = 0;
	if(ret == INVALID_LIT  && !bad_pol.empty()){ret = bad_pol[strt++];}

	while(strt<bad_pol.size()){var_act_hp.restoreElementIf(bad_pol[strt++].toVar());}

	return ret;
}*/

Literal BasicSolver::nextDecision(const bool rand){

	Variable var = INVALID_VAR;
    if(rand){
		//random decision
		do{
			if(var_act_hp.empty()) return INVALID_LIT;
			var     = var_act_hp.removeAtPos(glob.randy.nextInt(var_act_hp.size()));

		}while(vals.varKnwn(var) || glob.elim.isSet(var));

		cnt_stat(rnd_decs,1)

	}else{
		//activity based decision
		do{
			if(var_act_hp.empty()) return INVALID_LIT;
			var     = var_act_hp.removeTop();
		}while(vals.varKnwn(var) || glob.elim.isSet(var));

		cnt_stat(act_decs,1)
	}

	LOG2("Next decision chosen var " << var.toDIMACS());

	asrt(var != INVALID_VAR );

	//uint32_t pol_mode 	= glob.sets.pol_mode;//isExt(var) ? glob.sets.pol_mode_ext : glob.sets.pol_mode;
	bool pol 			= glob.sets.pol_mode  == 2 ? vals.varPrevValWasTrue(var) : glob.sets.pol_mode;
	return  var.toLit(pol);
}







int32_t	BasicSolver::bcp(){
	T_ON(prop);

    int32_t        	nbr_pivot_lits	= BCP_SUCCESS; //indicates no conflict by now
    uint32_t        num_props		= 0;

    while(queue_begin < assign_stck.size()){
        ++num_props;
        nbr_pivot_lits = bcpPropagateLit(assign_stck[queue_begin++].flipped());
    }

    cnt_stat(props,num_props)
    LOG2("Did " << num_props << " props");

    T_OFF(prop);
    return nbr_pivot_lits;
}





int32_t	BasicSolver::bcpPropagateLit(const Literal& false_lit){
	asrt(!glob.elim.isSet(false_lit.toVar())); LOG2("Propagate lit " << false_lit.flipped().toDIMACS());
	WIter<WtchEle> wi	= watching.getWtchOfW(false_lit);//watched clauses that have to be checked when lit gets false
	WtchEle wt(INVALID_LIT,true); WtchEle wcl = wt;
	const uint32_t dlev 	= decisionLevel();

	while(wi.getNext(wt)){

		//might be any other literal of this clause except the watched literal
        const Literal any_other = wt.getLit();   asrt(any_other != false_lit);

        if(vals.litTrue(any_other)){
        	wi.keep(wt);
        	if(!wt.isBinary()){	wi.keepNext();}
            continue;
        }

		if(wt.isBinary()){
			wi.keep(wt); asrt(!vals.litTrue(any_other));
			if(vals.litFalse(any_other)){
				bcpFinishOnConflict(false_lit,wi);
				int nbr_pivot_lits = bcpInitConflict(dlev,false_lit,any_other); 	asrt(nbr_pivot_lits != 1);
				return nbr_pivot_lits;
			}
			enqueue(any_other,false_lit);
			continue;
		}

        asrt(wi.hasNext());
        wi.getNext(wcl);//get actual clause pointer

        const Literal other_w = glob.getCls(wcl.getClsP()).otherWatcher(false_lit);

        //it might be that the shortcut literal in 'wt' does not point to the actual watched literal anymore
        if(vals.litTrue(other_w)){
        	wi.keep(WtchEle(other_w,false));
        	wi.keep(wcl);
            continue;
        }

        Cls& cl = glob.getCls(wcl.getClsP());
		if(watching.bcpFindNewWatcher(wcl.getClsP(),false_lit,other_w)){continue;}

		LOG2("No new watcher found for "<<false_lit.toDIMACS());
		// No new watched found -- clause is unit under assignment:
		wi.keep(WtchEle(other_w,false));
		wi.keep(wcl);

	    if (vals.litFalse(other_w)){

	    	bcpFinishOnConflict(false_lit,wi);

			int32_t  nbr_pivot_lits = bcpInitConflict(dlev,false_lit,other_w);
			ClsLit* clit 			= cl.uwBeg();
			do{firstUIPCheckLiteral(clit->getLit(),dlev,nbr_pivot_lits);}while((clit++)->has_next);

			asrt(nbr_pivot_lits != 1);
			return nbr_pivot_lits;

	    }else{
	       	LOG2("Enqueue Literal " << other_w.toDIMACS());
	    	enqueue(other_w, wcl.getClsP(),cl.isLrnt() ? cl.compLBD() : BDReason::LBD_MAX);
	    	asrt(!cl.isLrnt() || cl.compLBD() == ass_info.reasonLBD(other_w.toVar()));
	    }
	}
	LOG2("Finish propagation of " << false_lit.flipped().toDIMACS());
	watching.finish(false_lit,wi);
    return BCP_SUCCESS;
}

bool BasicSolver::firstUIP(int32_t nbr_pivot_elements){
	asrt(nbr_pivot_elements >= 0);
	glob.stats.cnt_confl++;

	if(nbr_pivot_elements == 0) return glob.setInvalid("Learned empty clause");

	T_ON(conflana);
    const uint32_t   dec_level           (decisionLevel());
    uint32_t        assign_stack_index  = assign_stck.size() ;

    asrt(nbr_pivot_elements != 1);
    BDReason 	r_info;
    Literal 	pivot_lit;

    while(true){

    	//skip all unseen assignments
    	//since there always exists a UIP the index cannot become negative
    	while (!var_flags.isSet(assign_stck[--assign_stack_index].toVar())){}
    	asrt(assign_stack_index >= 0);

    	pivot_lit                       = assign_stck[assign_stack_index]; asrt(vals.litTrue(pivot_lit));
    	//can be already cleared since this lit cannot appear in any remaining (previous) reason because literals
    	//in reason clauses are all false (except for the one it is the assign reason)
    	var_flags                       .clear(pivot_lit.toVar());

    	if(--nbr_pivot_elements == 0){
    		asrt(ass_info.litAssLev(pivot_lit) == dec_level);
    		asrt(pivot_lit != INVALID_LIT);
    		lemma[0]                = pivot_lit.flipped();
    		T_OFF(conflana);
    		return true;
    	}

    	const BDReason& r_info		= ass_info.varAssignReason(pivot_lit.toVar()); asrt(!r_info.isDecision());

    	if(r_info.isBinReason()){
    		firstUIPCheckLiteral(r_info.getBinReason(),dec_level,nbr_pivot_elements);
        	//settings.console_out->writeString("Short Reason: %d %d ",pivot_lit.toDIMACS(),confl_info.getShortFirstReason().toDIMACS());
    	}else{
    		asrt(r_info.isClsReason());

    		ClsP clp				= r_info.getClsReason();
    		Cls& cl                 = glob.getCls(clp); asrt(ass_info.varAssLev(cl.otherWatcher(pivot_lit).toVar())==dec_level);

    		//do for all except the pivot_lit
    	    firstUIPCheckLiteral(cl.otherWatcher(pivot_lit),dec_level,nbr_pivot_elements);

    	    ClsLit* clit 			= cl.uwBeg();
    	    do{firstUIPCheckLiteral(clit->getLit(),dec_level,nbr_pivot_elements);}while((clit++)->has_next);

    	}
	}
    return false;
}


uint32_t BasicSolver::backtrackLevel(){
    asrt(ass_info.litAssLev(lemma[0]) == decisionLevel());
    uint32_t lsz 	= lemma.size();
    if(lsz < 2) return 0;

    uint32_t max_idx = 1;
    while(--lsz){
    	uint32_t lev = ass_info.litAssLev(lemma[lsz]);
    	zeroLevel(lev); var_flags.carelessClear(lemma[lsz].toVar());
        if(lev > ass_info.litAssLev(lemma[max_idx])){max_idx = lsz;}
    }

    //such that the literals with maximum level are the watchers
    //this is important for backtracking!
    Literal highest		= lemma[max_idx];
    lemma[max_idx]     	= lemma[1];
    lemma[1]           	= highest;

    asrt(var_flags.allCleared()); sanity(confl_levs.allCleared());

    return ass_info.litAssLev(lemma[1]);
}




void     BasicSolver::reduceSelfImpl(){

    //asrt(c_db.checkAssignReasons());

    if(lemma.size() < 2){asrt(var_flags.allCleared()); return;}

    asrt(var_flags_to_clear.empty());
    //var_flags.clearAll();

    uint cp,idx;
    for(cp = idx = 1; idx != lemma.size(); ++idx){
        if(litIsNecessary(lemma[idx])) {lemma[cp++] = lemma[idx];}}

    lemma.resize(cp);

    for(cp = var_flags_to_clear.size();cp--;) {var_flags.carelessClear(var_flags_to_clear[cp]);}
    var_flags_to_clear.clear();
}


bool      BasicSolver::litIsNecessary(Literal lit){

    if(ass_info.varIsDecision(lit.toVar())) return true;

    const Variable start(lit.toVar());

    uint32_t unmark_top(var_flags_to_clear.size());
    asrt(tmp_literal_store.empty());

    while(true){

        asrt(!ass_info.varIsDecision(lit.toVar()));
        const BDReason& r_info = ass_info.varAssignReason(lit.toVar());

        bool is_nec;

        if(r_info.isBinReason()){
            is_nec = reduceSelfImpCheckLiteral(r_info.getBinReason());

        }else{
            asrt(r_info.isClsReason() );

            ClsP clp	= r_info.getClsReason();
            Cls& cl		= glob.getCls(clp);

            is_nec = reduceSelfImpCheckLiteral(cl.otherWatcher(lit));
            ClsLit* clit 			= cl.uwBeg();
            while(!is_nec){
            	is_nec = reduceSelfImpCheckLiteral(clit->getLit());
            	if(!(clit++)->has_next){break;}
            };
        }

        if(is_nec){
            for( ;unmark_top < var_flags_to_clear.size();++unmark_top)   {var_flags.clear(var_flags_to_clear[unmark_top]);}
            for(int i = tmp_literal_store.size();i--;)                   {var_flags.clear(tmp_literal_store[i].toVar());}
            tmp_literal_store.clear();
            return true;
        }


        if(tmp_literal_store.empty()){
           //statistics.total_lemma_shortening += 1;
            //shorten_lemma_to_unmark.clear();
            var_flags_to_clear.push_back(start);
            return false;
        }

        lit = tmp_literal_store.back(); tmp_literal_store.pop_back();
        var_flags_to_clear.push_back(lit.toVar());
    }
}


bool 	BasicSolver::propUnits(vector<Literal>& units){
	asrt(decisionLevel()==0);
	while(!units.empty()){
		Literal lit = units.back(); units.pop_back();
		if(vals.litTrue(lit) || glob.elim.isSet(lit.toVar())) continue;
		if(vals.litFalse(lit)) return false;
		LOG2("Enqueue unit " << lit.toDIMACS() << " after simpl");;
		decide(lit);
	}
	return bcp() == BCP_SUCCESS;
}


TrBool BasicSolver::simplify( ){

	asrt(decisionLevel() == 0);

	if (!glob.isValid() || bcp() != BCP_SUCCESS){   return TrBool(glob.setInvalid("Bad implication of unit clauses"));}

    if(simpl_next > glob.stats.cnt_props){return UNKNOWN_TrBool;}

    LOG2("Check for simpl: Props are " << glob.stats.cnt_props << " next simp is at " << simpl_next << " next big simp  " << big_simpl_next);

    if (big_simpl_next <= glob.stats.cnt_props){
    	inc_stat(simpl_big,1) T_ON(bigsimpl);

    	vector<Literal> units;
		SimplyFire simp(glob,vals,var_act_hp,units,var_flags);
		simp.initSimpl(watching,glob.sets.assym_prob,!did_big_simp,true);
		did_big_simp = true;

		if(glob.sets.equal_vars_chk && !simp.replaceSCCs()){ 						T_OFF(bigsimpl);
			return TrBool(glob.setInvalid("Bad SCC replace"));}

		if(!simp.elimAndSubsum(glob.sets.blockd_cls_chk)){ 							T_OFF(bigsimpl);
			return TrBool(glob.setInvalid("Bad Elim and Subsum test"));}


		if(glob.sets.assym_prob && !simp.assymChecks(glob.sets.blockd_cls_chk)){ 	T_OFF(bigsimpl);
			return TrBool(glob.setInvalid("Bad Assym checks"));}


		if(glob.sets.equal_vars_chk && !simp.replaceSCCs()){ 						T_OFF(bigsimpl);
			return TrBool(glob.setInvalid("Bad SCC replace after simp"));}


		simp.getSimplResult(watching);
		for(uint32_t i=0;i!=units.size();++i){
			Literal d = units[i];
			if(vals.litTrue(d)) continue;
			if(vals.litFalse(d)){													T_OFF(bigsimpl);
				return TrBool(glob.setInvalid("Bad Unit after Simplification"));}

			decide(d);
		}

		simpl_next		= glob.stats.cnt_props + glob.stats.litsCount() ;
		big_simpl_next	= simpl_next +  (glob.sets.simpl_freq*glob.stats.litsCount()); asrt(big_simpl_next != 0);
		T_OFF(bigsimpl);

		LOG2("Current props are " << glob.stats.cnt_props << " next simp is at " << simpl_next << " next big simp  " << big_simpl_next);

    }else if (assign_stck.size() != simplify_assign_idx){
    	inc_stat(simpl_sml,1)
    	cleanDB(max32);
    	simpl_next		= glob.stats.cnt_props + glob.stats.litsCount();
    }

   	sanity(watching.asrtWtchingOK());


    if (!glob.isValid() || bcp() != BCP_SUCCESS){   return TrBool(glob.setInvalid("Bad BCP after Simplification"));}
    simplify_assign_idx	= assign_stck.size();


    return UNKNOWN_TrBool;
}



void	BasicSolver::cleanDB( const uint32_t& max_lbd){
	FeedbackForm mem_fdbk;
	glob.c_info.registerFeedbackForm(mem_fdbk);
	cleanDB(max_lbd,mem_fdbk,0,0);
}

void	BasicSolver::cleanDB( const uint32_t& max_lbd,FeedbackForm& mem_fdbk, const uint32_t blkd_lits, const uint32_t blkd_cls){

	glob.clearLBDDistri();
	glob.stats.cnt_bin_lts	= 0;
	glob.stats.cnt_lcls		= blkd_cls;
	glob.stats.cnt_lng_lts	= blkd_lits;


	WtchEle wt(INVALID_LIT,true); WtchEle wcl = wt; uint32_t lbd;

	for(uint32_t l=0;l!=watching.litsCount();++l){
		Literal lit(l);
		const bool kwn 		= ass_info.litFix(lit);
		WIter<WtchEle> wi	= watching.getWtchOfW(lit);

		while(wi.getNext(wt)){
			if(wt.isBinary()){
				if(!kwn && !ass_info.litFix(wt.getLit())){ wi.keep(wt); cnt_stat(bin_lts,1); }
				else{asrt(vals.litTrue(lit)||vals.litTrue(wt.getLit()));}
				continue;
			}

		    asrt(wi.hasNext());
		    wi					.getNext(wcl);//get actual clause pointer
		    ClsP clp			= wcl.getClsP();
		    Cls& cls			= glob.getCls(clp);
		    const Literal ow	= cls.otherWatcher(lit);

		    //marked for deletion 	//glob.clsFree(wcl.getClsP());
		    if(ow==lit){							goto SKIP_CLS;}

		    if(lit<ow){ //see it the first time - check sat and remove false literals

		    	//marked from outside since cls is reason for curr assignment
		    	if(mem_fdbk.isBorder(clp)){
		    		if(cls.isLrnt()){glob.registrLBD(cls.compLBD());}
												goto KEEP_CLS;}

		    	if(kwn || ass_info.litFix(ow)){
		    		cls.unsetWatchers();			goto SKIP_CLS;} //mark for deletion and don't keep in this list

		    	if(cls.isLrnt()){
		    		if((lbd=cls.compLBD())>max_lbd){
		    			cls.unsetWatchers();		goto SKIP_CLS;}//mark for deletion and don't keep in this list
		    		glob.registrLBD(lbd);
		    	}

		    	{ClsLit* clit	= cls.uwBeg();
		    	ClsLit* cp		= clit;
		    	do{
		    		if(!ass_info.litFix(clit->getLit())){
		    			(cp++)->lit	= clit->lit;		   	continue;}
					if(vals.litTrue(clit->getLit())){
						cls.unsetWatchers(); 				goto SKIP_CLS;} //mark for deletion
					asrt(vals.litFalse(clit->getLit()));

		    	}while((clit++)->has_next); asrt(cp<=clit); asrt(cp>=cls.uwBeg());

		    	const uint32_t uw_cnt = cp - cls.uwBeg();
				//new cls is binary - delete this clause and keep binary
				if(uw_cnt==0){
					cnt_stat(bin_lts,2);
					cls		.unsetWatchers();
					wi		.keep(WtchEle(ow,true));
					watching.clsCreateBinPart(ow,lit);
					goto SKIP_CLS;
				}

				mem_fdbk.markBorder(clp); //start of clause
				mem_fdbk.markBorder(clp+uw_cnt); //last word of clause
				(--cp)->has_next = 0;

				cnt_stat(lng_lts,(uw_cnt+2));
			    cnt_stat(lcls,1);}
		    }

		    KEEP_CLS:;
		    wi.keep(WtchEle(ow,false));
		    wi.keep(wcl);

		    SKIP_CLS:;
		}
		watching.finish(lit,wi);
	}
	glob.c_info.submitFeedback(mem_fdbk);
}



void BasicSolver::learntsGC(){

	cnt_stat(gc,1)
	uint32_t dec_strt = decisionLevel() == 0 ? assign_stck.size(): dec_lev_end[0];
	asrt(dec_lev_end[0] <= assign_stck.size());

	FeedbackForm mem_fdbk;
	glob.c_info.registerFeedbackForm(mem_fdbk);
	uint32_t blkd_lits = 0;
	uint32_t blkd_cls = 0;

	//pay attention that no reasons are deleted
	for(uint32_t i=dec_strt;i!=assign_stck.size();++i){
		const BDReason& r	= ass_info.varAssignReason(assign_stck[i].toVar());
		if(r.isClsReason()){
			ClsP clp	= r.getClsReason();
			Cls& cls	= glob.getCls(clp);
			if(cls.isLrnt()){glob.unregistrLBD(cls.compLBD());} //for better selection of max_lbd
			ClsLit* clit	= cls.uwBeg(); 			mem_fdbk.markBorder(clp);
			do{++clp;}while((clit++)->has_next);	mem_fdbk.markBorder(clp);
			++blkd_cls;
			blkd_lits += (2 + (clit-cls.uwBeg()));
		}
	}


	uint32_t left	= 0;				uint32_t right		= BDReason::LBD_MAX;
	uint32_t left_c	=glob.lbd_distri[left]; 	uint32_t right_c	= glob.lbd_distri[right];
	while(left<right){
		if(left_c<=right_c){
			left_c 	+= glob.lbd_distri[++left];}
		else{
			right_c += glob.lbd_distri[--right];}
	}
	asrt(left==right);
	left = glob.stats.litsCount() > 150000000 ? 2 : (left - (left == BDReason::LBD_MAX));

	LOG1("Clean DB at level " << decisionLevel() << " with max_lbd "<<left);

	cleanDB(left,mem_fdbk,blkd_lits,blkd_cls);

	glob.sets.max_learnts	+= glob.sets.max_learnts_inc;
}



bool BasicSolver::parseDIMACS(IFStream& in) {
	vector<Literal> parsed_lits;
	int32_t  parsed_lit;
	Literal lit; //uint32_t c = 0; uint64_t l = 0; uint32_t lts = 0;

	while(in.hasNext()){
		if(in.currEquals('c')) {in.skipLine();}
		else if(in.currEquals('p')){
			if (in == "p cnf"){
				reserveForVariables(in.getNextInt());
				reserveForClauses(in.getNextInt());
				in.skipLine();
			}else{ cout << "c PARSE ERROR in file " << glob.sets.cnf_file_name << endl, exit(EXIT_FAILURE); }
		//}else if(in.currEquals('s')){
			//should = (in == "s SATISFIABLE") ? TRUE_TrBool : FALSE_TrBool;
		}else{

			while((parsed_lit = in.getNextInt()) != 0){
					lit = Literal::DIMACS2Literal(parsed_lit); asrt(parsed_lit == lit.toDIMACS());
					parsed_lits.push_back(lit);
			}
			if(!addClause(parsed_lits)){
				LOG1("Formula contains an empty clause!! - is unsatisfiable!\n");
				return glob.setInvalid("Empty Clause in parsed formula");
			}
			parsed_lits.clear();
		}
	}
	return true;
}


void		BasicSolver::reinsertElimClses(){
	asrt(glob.isValid()); LOG1("Start reinsertion of eliminated clauses at level " << decisionLevel());
	glob.elim_clses.initClsPopping();
	Literal	*it, *end, *lits, first; ElimType type;

	while(glob.elim_clses.popCls(lits,end,type)){ asrt(lits<end);
		first = *lits;
		LOG2("Next element in reinsertion start with " << first.toDIMACS());

		if(type==VAR_ENSURE_VAL){
			glob.elim.clear(first.toVar());
			LOG2("Reinsertion ensure val of elim " << first.toDIMACS());
			if(vals.litUkwn(first)){
				LOG2("Reinsertion decides on variable " << first.toDIMACS());
				decide(first);		aux_var(int32_t bcp_res =)
				bcp(); 				asrt(bcp_res==BCP_SUCCESS);}
			continue;
		}

		it		= lits;
		for(;it!=end;++it){
			if(vals.litTrue(*it)){LOG2("Cls is true by lit " << it->toDIMACS());
				goto CHECK_NEXT;} //continue with next clause
			if(vals.litUkwn(*it)){asrt(first==*lits); LOG2("Unknown lit in cls " << it->toDIMACS());
				*lits=*it; *it=first; }
			else{asrt(vals.litFalse(*it)); asrt(!glob.isElim(it->toVar())); LOG2("False lit in cls " << it->toDIMACS());}
		}

		if(vals.litFalse(*lits)){
			LOG1("Reinsertion of blocked clause forces variable flip of " << lits->toDIMACS()); asrt(!glob.isElim(lits->toVar()));asrt(type==BLKD_CLS);
			vals	.varClearValue(lits->toVar());
		}

		asrt(vals.litUkwn(*lits));
		glob.elim.clear(lits->toVar()); //may be set or not
		{decide(*lits);		aux_var(int32_t bcp_res =)
		bcp(); 						asrt(bcp_res==BCP_SUCCESS);

		const uint32_t sz = end - lits;
		if(sz==2){watching.clsCreateBinary(*lits,*(lits+1));}
		else if(sz>2){	clsCreate(lits,sz,false,1);}}

		CHECK_NEXT:;
	}
	LOG1("Finish reinsertion of eliminated clauses");
}

/*
void		BasicSolver::writeDIMACS(const char* file,bool print_elim, bool print_lrnt){
	std::ofstream out(file); asrt(!assign_stck.empty()); asrt(assign_stck[0]==TRUE_LIT);
	out << "p cnf " << var_act_comp.ext_border.idx()-1 << " ";
	long cls_p = out.tellp(); out << "              " << endl;

	uint32_t clses = assign_stck.size();
	for(uint32_t i=1;i<assign_stck.size();++i){
		out << assign_stck[i].toDIMACS() << " 0" << endl;}

	clses += watching.writeDimacs(out,print_lrnt);

	if(print_elim){
		elim_clses.initClsPopping();
		Literal	*it, *end; ElimType type;
		bool first = true;

		while(elim_clses.popCls(it,end,type)){ asrt(it<end);
			if(type==VAR_ENSURE_VAL){ continue;}
			if(first){
				first = false;
				out << endl << "c start eliminated clauses" << endl;
			}
			for(;it!=end;++it){	out << it->toDIMACS() << " ";}
			out << "0" << endl;
			++clses;
		}
	}
	out.seekp(cls_p);
	out << clses;
}*/

/*

uint32_t	BasicSolver::findNextAssStckExtCand(uint32_t idx){
	while(idx--){
		Literal lit = assign_stck[idx];
		if(ass_info.litFix(lit)) 					return 0;
		if(!ass_info.varIsShortReason(lit.toVar())) 	return idx;
	}
	return 0;
}


void 	BasicSolver::applyAssStckVarExt(){
	if(decisionLevel()==0 || glob.sets.max_ext_vars_ass ==0) return;

	uint32_t r_idx	= findNextAssStckExtCand(assign_stck.size()); //find highest assignment without bin reason
	Literal ncls[3];

	while(r_idx && glob.sets.max_ext_vars_ass){
		uint32_t l_idx	= findNextAssStckExtCand(r_idx); asrt(l_idx < r_idx); //find second highest assignment without bin reason
		if(l_idx==0)	return;
		ncls[0]		= assign_stck[l_idx]; Variable a = ncls[0].toVar();
		ncls[1]		= assign_stck[r_idx]; Variable b = ncls[1].toVar();

		if(containsExtWith(a,b) || containsExtWith(b,a)){ //extended var for relation of these to lits already exists
			r_idx = l_idx; continue; //forget about highest assignment and continue with sec highest ...
		}
		//introduce extended var

		//a b x
		ncls[2] 		= getNewExtendedVar().toPosLit();
		vals			.litSetPrevVal(ncls[2]); // set default value to true
		//cout << "is " << ncls[0].toDIMACS() << " " << ncls[1].toDIMACS() << " " << ncls[2].toDIMACS() << endl;
		clsCreate(ncls,3,false,3);

		//-a -b x
		ncls[0]		= ncls[0].flipped();
		ncls[1]		= ncls[1].flipped();
		//cout << "is " << ncls[0].toDIMACS() << " " << ncls[1].toDIMACS() << " " << ncls[2].toDIMACS() << endl;
		clsCreate(ncls,3,false,3);

		//a -b -x
		ncls[0]		= ncls[0].flipped();
		ncls[2]		= ncls[2].flipped();
		//cout << "is " << ncls[0].toDIMACS() << " " << ncls[1].toDIMACS() << " " << ncls[2].toDIMACS() << endl;
		clsCreate(ncls,3,false,3);

		//-a b -x
		ncls[0]		= ncls[0].flipped();
		ncls[1]		= ncls[1].flipped();
		//cout << "is " << ncls[0].toDIMACS() << " " << ncls[1].toDIMACS() << " " << ncls[2].toDIMACS() << endl;
		clsCreate(ncls,3,false,3);

		r_idx 	= findNextAssStckExtCand(l_idx); //continue with next two assignments
		glob.sets		.max_ext_vars_ass--;
		stats	.nbr_ext_vars_ass++;
		LOG_MEDIUM("Create ext var for " << a.toDIMACS() << " and " << b.toDIMACS());
	}


}

void		BasicSolver::collectKnownExtVars(){
	asrt(decisionLevel()==0);
	cleanDB(true,max32,false);

	uint32_t cp = 1;
	for(uint32_t i=1;i!=assign_stck.size();++i){
		Variable v = assign_stck[i].toVar(); asrt(vals.varKnwn(v));
		if(isExt(v)){
			free_extended		.push_back(v);
			vals				.varClearValue(v);
		}else{
			assign_stck[cp++]	=assign_stck[i];
		}
	}
	assign_stck.resize(cp);
}

bool		BasicSolver::containsExtWith(const Variable& who, const Variable& with){
	Literal l			= watching.getLitWithSmallerList(who);
	Iter<WtchEle> it 	= watching.getWtchOf(l);
	WtchEle we(FALSE_LIT,true);
	while(it.getNext(we)){
		if(we.isBinary()){ continue;} asrt(it.hasNext());

		it			.getNext(we);
		Cls& cls 	= getCls(we.getClsP());
		if(cls.unWatchdCnt() != 1) continue;

		Variable a = cls.otherWatcher(l).toVar();
		Variable b = cls.uwBeg()->toVar();

		if(	(isExt(a) && b == with)
		||	(isExt(b) && a == with)) return true;
	}
	return false;
}

Variable BasicSolver::getNewExtendedVar(){
	Variable ext  = INVALID_VAR;
	if(free_extended.empty()){
		ext			= next_ext_var;
		next_ext_var	= Variable(ext.idx()+1);
	}else{
		ext			= free_extended.back();
		free_extended	.pop_back();
	}asrt(ext != INVALID_VAR);

	reserveForVariables(ext.idx());
	asrt(!vals.varKnwn(ext));

	return ext;
}
*/


void		BasicSolver::printModel(){
	cout << endl << "v "; Variable max (ass_info.varsCount());
	for(Variable v(1);v<max;v=Variable(v.idx()+1)){
	//for(Variable v(1);v<var_act_comp.ext_border;v=Variable(v.idx()+1)){
		asrt(vals.varKnwn(v));
		cout << v.toLit(vals.varPrevValWasTrue(v)).toDIMACS() << " " ;
	}
	cout << " 0" << endl;
}

bool		BasicSolver::asrtAllFalse(Literal* from, Literal* end){
	if(from>end) return false;
	while(from!=end){
		if(!vals.litFalse(*from)) return false;
	}
	return true;
}


bool BasicSolver::asrtVerifyModel(const char* file) {
	IFStream in(file);
	if(in.isBad()){
		LOG0("Could not read file to validate Model");
		return false;
	}

	vector<Literal> cls;
	int32_t  parsed_lit; bool ret = true;
	Literal lit; //uint32_t c = 0; uint64_t l = 0; uint32_t lts = 0;

	while(in.hasNext()){
		if(in.currEquals('c')||in.currEquals('p')) {in.skipLine();}
		else{
			bool sat = false;
			while((parsed_lit = in.getNextInt()) != 0){
				lit 	= Literal::DIMACS2Literal(parsed_lit); asrt(parsed_lit == lit.toDIMACS());
				sat 	|= vals.litTrue(lit);
				cls	.push_back(lit);
			}
			if(!sat){
				LOG0("c Model does not sat cls: ");
				for(uint32_t i=0;i!=cls.size();++i){LOG0(cls[i].toDIMACS());}
				ret = false;
			}
			cls.clear();
		}
	}
	LOG0("Verification of model is " << (ret?"supi dupi":"bad"));
	return ret;;
}






/*
void BasicSolver::clearBinWatchers(Literal lit){

	asrt(v_info.vals.litUkwn(lit));
	we_iter from,to,end;
	we_vec& wes = litWatchedClauses(lit);

	from = to = wes.begin(); end = wes.end();
	while(from != end){
		WatchedElement& we = *from++;
		if(we.isBinary() && v_info.vals.litTrue(we.getLiteral())) {continue;} //delete watcher

		*to++ = we;
	}
	if(to != end){wes.erase(to,end);}
}



void BasicSolver::clearBinWatchers(Variable var){
	if(var_flags.isSet(var.idx()) || v_info.vals.varKnwn(var)) return;
	var_flags.setFlag(var.idx());	var_flags_to_clear.push_back(var);

	clearBinWatchers(var.toNegLit());
	clearBinWatchers(var.toPosLit());
}*/

/*
void	BasicSolver::simplifyBinWatchersOf (we_vec& wes){
	we_iter from,to;

	for(from = to = wes.begin(); from != wes.end();++from){
		if(from->isBinary()){ clearBinWatchers(from->getLiteral().toVar());	}
		else{*to++ = *from;}		//might be since bigsimplify changes are not simplified further
	}
	wes.erase(to,wes.end());
}



void	BasicSolver::simplifyAllBinWatchers (){
	asrt(var_flags.allCleared());
	asrt(decisionLevel() == 0);

	while(simplify_assign_idx < assign_stack.size()){
		Literal lit = assign_stack[simplify_assign_idx++];
		simplifyBinWatchersOf(litWatchedClauses(lit));
		simplifyBinWatchersOf(litWatchedClauses(lit.flipped()));
	}


	for(uint32_t i = var_flags_to_clear.size();i--;){var_flags.carelessClear(var_flags_to_clear[i].idx());}
	var_flags_to_clear.clear();
	asrt(var_flags.allCleared());
}*/
