#ifndef BASICSOLVER_HPP_
#define BASICSOLVER_HPP_
#define basicsolver_hpp_version "$Revision: 1.19 $"

#include    <vector>

#include "ThreadGlobals.hpp"
#include 	"Assertion.hpp"
#include	"Types.hpp"
#include "Settings.hpp"
#include	"LazyFraction.hpp"
#include "Satistics.hpp"
#include "Heap.hpp"
#include "BitVector.hpp"
#include	"AssignInfos.hpp"
#include 	"ConflictsRestarts.hpp"
#include	"MemPoolA.hpp"
#include	"VecOfLists.hpp"
#include 	"Hash.hpp"
#include 	"IFStream.hpp"
#include 	"Logging.hpp"
#include	"Watching.hpp"
#include	"HistoryBuffer.hpp"

using        std::vector;


typedef vector<ClsP> ::iterator cls_iter_t;




class BasicSolver{

protected:
	static const int32_t BCP_SUCCESS 	= -1; //SOMETHING < 0
	TGlobals&							glob;

    /** A heuristic measurement of the activity of a variable  */
	//Variable								next_ext_var;
	VarActComp							var_act_comp;
    var_hp_t   							var_act_hp;

    vector<Literal>						assign_stck;
public:
    VarVal								vals;
    Assigns								ass_info;
    Watching							watching;


    /** Head of queue - as index into the assign_stack */
    uint32_t                        	queue_begin;

    /** Indices for different decision levels in 'assign_stack' */
    vector<uint32_t>                	dec_lev_end;

    /** Number of top-level assignments at last execution of 'simplify()' */
    uint32_t                        simplify_assign_idx;

    /** Remaining number of propagations that must be made before next execution of 'simplify()' */
    uint64_t                         simpl_next;
    uint64_t                         big_simpl_next;
    bool							 did_big_simp;

    /** Data structure to set a flag for a variable and allow efficient clearing of all flags */
    BitVector<Variable>	            var_flags;

    LubyRestarts					luby_restarts;
    HistoryBuffer<uint8_t,uint64_t> lbd_history;
    uint64_t						lbd_total;

    vector<Variable>				conflVars;

    //conflict related information
    vector<Literal>					lemma;
    BitVector<uint32_t>				confl_levs;
    uint32_t						confl_lbd;



    bool setLevel(const uint32_t& l){asrt(l<=confl_levs.size()); return confl_levs.setFlagNew(l);}
    void zeroLevel(const uint32_t& l){asrt(l<=confl_levs.size()); confl_levs.carelessClear(l);}
    bool isSetLevel(const uint32_t& l){asrt(l<=confl_levs.size()); return confl_levs.isSet(l);}


    vector<Variable>                var_flags_to_clear;
    vector<Literal>                 tmp_literal_store;


//internal Methods
    inline void     newDecisionLevel    ();
    inline void     enqueue(const Literal& lit, const Literal& from);
    inline void     enqueue(const Literal& lit, const ClsP& from, const uint32_t& lbd);
    inline void     decide(const Literal& lit);
    inline void     varDecayActivity();
    inline uint32_t  decisionLevel()  const;
    inline uint32_t  assignsAtLevel0 ()const;

    void 			clearBinWatchers(Variable var);
    void 			clearBinWatchers(Literal lit);
    TrBool          simplify            ( );
    bool			propUnits			(vector<Literal>& units);
    void			cleanDB				( const uint32_t& max_lbd, FeedbackForm& mem_fdbk, const uint32_t blkd_lits, const uint32_t blkd_cls);
    void			cleanDB				( const uint32_t& max_lbd);
    void 			learntsGC			();
    TrBool			preProcess();

    int32_t        	bcp           		( );
    int32_t        	bcpPropagateLit		(const Literal& false_lit);
    int32_t			bcpBins				(const Literal& false_lit);
    inline void 	bcpFinishOnConflict	(const Literal& lit, WIter<WtchEle>& iter);
    inline int32_t	bcpInitConflict		(const int32_t& dec_level, const Literal& w1, const Literal& w2);

    inline void    	backjumpKeepLevel   (uint32_t level);    // Backtrack until a certain level.
    void    		backjumpUndoAssigns (uint32_t assign_idx); //undoes all assignments above and including the specified assign_idx
    bool        	firstUIP            (int32_t nbr_pivot_lits);
    inline void     firstUIPCheckLiteral(const Literal& lit,const int32_t& dec_level,int32_t& nbr_pivot_elements);
    uint32_t        backtrackLevel      ();
    void            reduceSelfImpl  	();
    //inline void		unsetConfLemmaFlags	();
    bool            litIsNecessary      (Literal lit);
    inline bool     reduceSelfImpCheckLiteral (const Literal& lit);

    inline void		varBump				(const Variable& var);
    TrBool          search             	(uint32_t max_confl);
    Literal         nextDecision(const bool rand );
    Literal         nextDecision(){
					return nextDecision(ass_info.lessAgilThan(1000000) && glob.randy.nextInt(1000) < glob.sets.random_var_freq_permil); }
   // Literal         nextDecisionStrict();
    bool            uncheckedAddClause  (vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend);

   /* struct LitActDescendingFunctor{
    	BasicSolver* outer ;
    	LitActDescendingFunctor(BasicSolver* s):outer(s){};
      	bool operator() (const Literal& a, const Literal& b){return outer->var_act_hp.getValue(a.toVar()) > outer->var_act_hp.getValue(b.toVar());	}
    };*/

    ClsP			clsCreate(Literal* lbeg, uint32_t sz,const bool learnt ,const uint32_t& lbd);

    //vector<Variable> free_extended;
    //bool			isExt(const Variable& v) const{return var_act_comp.ext_border.idx() <= v.idx();}
    Variable		getNewExtendedVar();
    void			collectKnownExtVars();
    uint32_t		findNextAssStckExtCand(uint32_t idx);
    void			applyAssStckVarExt();
    bool			containsExtWith(const Variable& who, const Variable& with);



public:
                    BasicSolver(TGlobals& g);
                    ~BasicSolver();

    inline void   	reserveForVariables(uint32_t max_idx);
    inline void 	reserveForClauses(uint32_t need_sz);
	bool    		addClause(vector<Literal>& lits);
	void			reinsertElimClses();

	TrBool 		isSAT();
	bool 		parseDIMACS(IFStream& in);
	void			writeDIMACS(const char* file,bool print_elim, bool print_lrnt);
	void			printModel();
private:
	bool			asrtAllFalse(Literal* from, Literal* end);
	bool			asrtNoBadLevels();
public:
	bool 		asrtVerifyModel(const char* file);

};



//Inline Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


void		BasicSolver::varBump(const Variable& var){
	var_act_hp          .activityBump(var);
}

void    BasicSolver::enqueue(const Literal& lit, const Literal& from){
    asrt(vals.litUkwn(lit));
    ass_info.setValueReason(lit,decisionLevel(),from); asrt(ass_info.reasonLBD(lit.toVar()) == BDReason::LBD_MAX);
    assign_stck.push_back(lit);
}

void     BasicSolver::enqueue(const Literal& lit, const ClsP& from, const uint32_t& lbd){
	asrt(vals.litUkwn(lit));
	ass_info.setValueReason(lit,decisionLevel(),from,lbd); asrt(ass_info.reasonLBD(lit.toVar()) == lbd);
	assign_stck.push_back(lit);
}

void    BasicSolver::decide(const Literal& lit){
    asrt(vals.litUkwn(lit));
    ass_info.setValueDecision(lit,decisionLevel()); asrt(ass_info.reasonLBD(lit.toVar()) == BDReason::LBD_MAX);
    assign_stck.push_back(lit);
}



uint32_t    BasicSolver::decisionLevel ()			const	{return dec_lev_end.size(); }
uint32_t    BasicSolver::assignsAtLevel0 ()		const	{return dec_lev_end.empty() ? assign_stck.size(): dec_lev_end[0]; }

void        BasicSolver::newDecisionLevel()                	{
			dec_lev_end.push_back(assign_stck.size()); sanity(confl_levs.allCleared());
			confl_levs.ensureSize(dec_lev_end.size()+1,false);}



void     	BasicSolver::reserveForVariables(uint32_t max_idx){

	if(++max_idx > glob.elim.size()){
	    LOG2("reserve for variables " << max_idx);

	    vals			.ensureVarVals(max_idx);
	    ass_info		.ensureVars(max_idx);
	    var_act_hp	.insertElementsTo(max_idx,glob.sets.var_min_inc);
	    var_flags		.ensureSize(max_idx,false);
	    glob			.ensureVarVals(max_idx);

	    max_idx <<= 1;
	    watching	.enlarge(max_idx);
	 }
}

void BasicSolver::reserveForClauses(uint32_t need_sz){
	static const uint32_t avg_cls_len = 2;

	glob.c_info	.reserveFor(need_sz * avg_cls_len);
	watching		.reserveForTotal(need_sz * 2);
}

void    	BasicSolver::backjumpKeepLevel(uint32_t level){

    if(decisionLevel() <= level) return;
    backjumpUndoAssigns(dec_lev_end[level]);

    dec_lev_end    	.resize(level);
}


void BasicSolver::bcpFinishOnConflict(const Literal& lit, WIter<WtchEle>& iter){
	LOG2("Finish on conflict with "<<lit.toDIMACS());
	queue_begin 	= assign_stck.size(); // to finish outer while loop
	watching		.keepRest(lit,iter);
}


int32_t	BasicSolver::bcpInitConflict(const int32_t& dec_level, const Literal& w1, const Literal& w2){
	asrt(var_flags.allCleared()); LOG2("init conflict with "<<w1.toDIMACS()<<" and " << w2.toDIMACS());

	lemma					.resize(1,INVALID_LIT); sanity(confl_levs.allCleared());
	confl_lbd 				= 1;
	conflVars				.clear();
	int32_t nbr_pivot_lits 	= 0;
	firstUIPCheckLiteral(w1,dec_level,nbr_pivot_lits);
	firstUIPCheckLiteral(w2,dec_level,nbr_pivot_lits);
	return nbr_pivot_lits;
}



void BasicSolver::firstUIPCheckLiteral(const Literal& lit,const int32_t& dec_level,int& nbr_pivot_elements){
   const Variable      var(lit.toVar());

   if(var_flags.isSet(var)) return;
   asrt(vals.litFalse(lit));

   int level = ass_info.varAssLev(var);
   if(level > 0){
       var_flags               .setFlag(var);
       varBump(var);			LOG2("Bump var " << var.toDIMACS());

       if(level <  dec_level){
           lemma			.push_back(lit);
           confl_lbd		+= setLevel(level);
       }else {
           ++nbr_pivot_elements;

           //glucose
           //if(ass_info.varAssignReason(var).getLBD() != BDReason::LBD_MAX){
           if(glob.sets.bump_glucose){conflVars.push_back(var);}//}
       }
   }
}


bool BasicSolver::reduceSelfImpCheckLiteral(const Literal& lit){

    Variable var = lit.toVar();   asrt(var.idx() < var_flags.size());

    if(var_flags.isSet(var)) return false;

    const uint32_t lev	= ass_info.varAssLev(var);
    if(lev > 0){
        if(ass_info.varIsDecision(var) || !isSetLevel(lev)){
            return true;}

        var_flags			.setFlag(var);
        tmp_literal_store	.push_back(lit);
    }
    return false;
}





const char*  basicsolver_cpp_version();
const char*  main_cpp_version();
#endif /*BASICSOLVER_HPP_*/
