
#include "ThreadGlobals.hpp"
#include "Satistics.hpp"
#include <signal.h>

#include "Types.hpp"
#include "IFStream.hpp"
#include "Logging.hpp"
#include "BasicSolver.hpp"
#include <string>
#include <algorithm>
#include <cctype>
//#include "DMRPSolver.hpp"

const char*  main_cpp_version(){return "$Revision: 1.14 $";}


/**
 * Some checks
 */

bool check_defines(){

	std::cout << "c asrts will be evaluated and wordsize is " << sizeof(void*) << std::endl;
    return
    //sizeof(void*)           ==      32/8                &&
    sizeof(1)				==		32/8				&&
    sizeof(int32_t)         ==      32/8                &&
    sizeof(int64_t)         ==      64/8                &&
    sizeof(uint32_t)        ==      32/8                &&
    sizeof(uint64_t)        ==      64/8                &&
    sizeof(double)          ==      64/8                &&
    sizeof(float)           ==      32/8                &&
    sizeof(Literal)         ==      sizeof(uint32_t)    &&
    sizeof(Variable)        ==      sizeof(uint32_t)    &&
    sizeof(float)           ==      sizeof(uint32_t)    &&
    ((int)false)            ==      0                   &&
    ((int)true)             ==      1;
}



LOGGING_INIT("./log_solver.csv");
ASRT_INIT

TGlobals					glob;

static void SIGINT_handler(int signum) {
	string what="x";
	do{
		std::transform(what.begin(), what.end(), what.begin(),tolower);
		if(what=="k")	{ exit(0);}
		if(what=="q")	{ glob.sets.timeout = 0; break;}
		if(what=="p")	{ glob.stats.printStatsConsole(UNKNOWN_TrBool);}
		if(what=="c")	{ break;}
		if(what=="o")	{ glob.sets.printUsage();}
		if(what=="r")	{ std::cin >> what; glob.sets.read(what.c_str());}


		cout << endl << "What to do?" << endl;
		cout << "k:       kill immediately" << endl;
		cout << "q:       quit after next restart" << endl;
		cout << "p:       print statistics" << endl;
		cout << "o:       show options and current option setting" << endl;
		cout << "r <opt>: read (non initial) option with new value (e.g. r -pol-mode=1)" << endl;
		cout << "c:       continue" << endl;
	}while(std::cin >> what);
	cout << "go ahead ... " << endl;
}




int main(int argc, char** argv){

	BasicSolver		solver(glob);
	Settings& sets	= glob.sets;
	Satistics& stats	= glob.stats;

	sanity((printf("sanity checks will be evaluated\n"),true));
    asrt(check_defines()); asrt(solver.vals.litTrue(TRUE_LIT));

    //LOGGING_OFF ASRT_OFF

    if(!sets.read(argc,argv)){
        sets.printUsage();
        return EXIT_FAILURE;
    }
    glob.randy.setSeed(sets.rand_seed);

    //try to open file
    IFStream in(sets.cnf_file_name.c_str()); asrt(solver.vals.litTrue(TRUE_LIT));
    //FILE * in = fopen (sets.cnf_file_name, "r");
    if(in.isBad()){
        cout << "c Could not open CNF file " <<  sets.cnf_file_name << endl;
        return EXIT_FAILURE;
    }

    signal(SIGINT,SIGINT_handler);
    //signal(SIGQUIT,SIGQUIT_handler);

    if(sets.print_console ){stats.printProgressHead();}

    T_ON(parse); asrt(solver.vals.litTrue(TRUE_LIT));
    bool ok=solver.parseDIMACS(in);
    in.close();
    T_OFF(parse);

    glob.stats.keepOrig(glob.elim.size());

    asrt(solver.vals.litTrue(TRUE_LIT));
    if(ok&&sets.print_console ){stats.printProgress();}

    //T_ON(eq_but_one);
    //solver.findOneDiffClses();
    //T_OFF(eq_but_one);

    asrt(solver.vals.litTrue(TRUE_LIT));
    TrBool result = ok ? solver.isSAT() : FALSE_TrBool;
    //TrBool result = ok  ? UNKNOWN_TrBool : FALSE_TrBool;
    //TrBool result = ok && solver.subsumptionTest() ? UNKNOWN_TrBool : FALSE_TrBool;
    //TrBool result = ok && solver.shrinkyWinky() ? UNKNOWN_TrBool : FALSE_TrBool;

    //char* result_txt = (char*)"UNKNOWN";

    if(result.isKnown()){
		if(result.isTrue()){

			cout << endl << "s SATISFIABLE" << endl;
			solver.reinsertElimClses();
			if(sets.print_model) {solver.printModel();}

			/*if(!solver.asrtVerifyModel(sets.cnf_file_name.c_str())){
				LOG0("Model could not be validated");}*/
		}else{
			cout << endl << "s UNSATISFIABLE" << endl;
			char* why =  glob.whyInvalid();
			if(why != NULL){
				cout << "c Reason for Unsat: " << why << endl;}

		}
   }else{
	   printf("s UNKNOWN\n");
   }

    if(!sets.csv_file_name.empty()){
        OFStream out(sets.csv_file_name.c_str(),false);
        stats.printCSV(sets.csv_file_name.c_str(),result);}


    if(sets.print_console){
    	stats.printStatsConsole(result);
    }

    /*if(!sets.dimacs_file_name.empty()){
    	solver.writeDIMACS(sets.dimacs_file_name.c_str(),true,false);}*/

    /*
    uint64_t ct = c_info.getMemTotal();
    uint64_t wt = solver.watches.getMemTotal() ;
    uint64_t vt	= v_vals.getMemTotal();
    uint64_t et = elim.getMemTotal();

    cout << "Mem total of c_info is " << ct << " byte is mega " << (ct/1000000) << endl;
    cout << "Mem total of watches is " <<wt << " byte is mega " << (wt/1000000) << endl;
    cout << "Mem total of vals is " <<vt << " byte is mega " << (vt/1000000) << endl;
    cout << "Mem total of elim is " <<et << " byte is mega " << (et/1000000) << endl;*/

    uint32_t ret_val = result.isKnown() ? (result.isTrue()?10:20) : 0;
#ifdef NDEBUG
	exit(ret_val);
#endif
    return ret_val;
}







