/*
 * Hash.hpp
 *
 *  Created on: 22.09.2010
 *      Author: stephan
 */

#ifndef HASH_HPP_
#define HASH_HPP_
#define hash_hpp_version	  "$Revision: 1.2 $"

#include <stdint.h>

inline uint32_t hashBit32(const int& x)                    { return 1 << (x & 31);}

inline uint64_t hashBit64(const int& x)                    { return 1 << (x & 63);}

inline uint64_t SDBMHash(uint32_t* vals, uint32_t sz,uint32_t skip_idx = ~0){
	uint64_t hash = 0;
	while(sz--){
		if(sz==skip_idx) continue;
		hash = vals[sz] + (hash << 6) + (hash << 16) - hash;}

	return hash;
}

template<class I>
inline uint64_t SDBMHash(I* vals, uint32_t sz,uint32_t skip_idx = ~0){
	uint64_t hash = 0;
	while(sz--){
		if(sz==skip_idx) continue;
		hash = vals[sz].idx() + (hash << 6) + (hash << 16) - hash;}

	return hash;
}


inline uint64_t SDBMHash(uint32_t val){
	static const uint32_t gran 	= 4;
	//static const uint32_t mask	= (2^gran) - 1;

	uint64_t hash = 0;
	for(uint32_t i=(sizeof(val)*8)/gran;i--;){
		//hash = (val&mask) + (hash << 6) + (hash << 16) - hash;
		hash = val + (hash << 6) + (hash << 16) - hash;
		val	>>= gran;
	}
	asrt(val == 0);

	return hash;
}


//number of set bits in an integer
class BinCrossSum{
	uint8_t csum[256];

public:
	inline BinCrossSum();

	inline uint8_t get8(const uint8_t& x);
	inline uint8_t get16(const uint16_t& x);
	inline uint8_t get32(const uint32_t& x);
	inline uint8_t get64(const uint64_t& x);
};

BinCrossSum::BinCrossSum(){
	uint32_t j 	= 0;
	csum[0] 	= 0;

	for(uint32_t i=1;i!=256;++i){

		if(j<<1 == i) j = 0;

		csum[i] = 1 + csum[j++];
	}
	asrt(csum[255]==8);
}

uint8_t BinCrossSum::get8(const uint8_t& x){
	return csum[x];
}

uint8_t BinCrossSum::get16(const uint16_t& x){
	return get8(x) + get8(x>>8);
}

uint8_t BinCrossSum::get32(const uint32_t& x){
	return get16(x) + get16(x>>16);
}


uint8_t BinCrossSum::get64(const uint64_t& x){
	return get32(x) + get32(x>>32);
}

inline uint64_t set_all(const bool b){
	return ((uint64_t)(!b)) - 1;
}


#endif /* HASH_HPP_ */
